@@ -1,11 +1,10 @@
 import pickle
-from contextlib import contextmanager
-from copy import deepcopy
+from copy import _reconstruct, deepcopy  # type: ignore[attr-defined]
 from functools import cached_property
 from inspect import getmembers
+from itertools import chain
 from pickletools import optimize
-from typing import Any, Dict, Iterator, Optional, Tuple, TypeVar, Union
-from warnings import warn
+from typing import Any, Dict, Iterable, List, Optional, Tuple, TypeVar, Union
 
 
 def is_casadi_object(obj: Any) -> bool:
@@ -50,67 +49,35 @@ def is_pickleable(obj: Any) -> bool:
         return False
 
 
-def _get_dict_state(obj: Any) -> Dict[str, Any]:
+def _get_dict_state(
+    obj: "SupportsDeepcopyAndPickle",
+    attributes: Iterable[str],
+    fullstate: bool,
+    remove_None: bool,
+) -> Dict[str, Any]:
     """Internal utility for SupportsDeepcopyAndPickle."""
-    dictstate: Dict = obj.__dict__.copy()
-    dictstate["_GETFULLSTATE"] = None
-    if not obj._GETFULLSTATE:
-        warn(
-            f"to pickle {obj.__class__.__name__}, all references to CasADi and "
-            "unpickleable objects are removed from __dict__.",
-            RuntimeWarning,
-        )
-        for attr, val in obj.__dict__.items():
-            if is_casadi_object(val) or not is_pickleable(val):
-                dictstate.pop(attr, None)
-    return dictstate
-
-
-def _get_slots_state(obj: Any) -> Dict[str, Any]:
-    """Internal utility for SupportsDeepcopyAndPickle."""
-    slotsstate: Dict[str, Any] = {n: getattr(obj, n, None) for n in obj.__slots__}
-    if not obj._GETFULLSTATE:
-        warn(
-            f"to pickle {obj.__class__.__name__}, all references to CasADi and "
-            "unpickleable objects are removed from __slots__.",
-            RuntimeWarning,
-        )
-        for attr in obj.__slots__:
-            val = slotsstate[attr]
-            if val is None or is_casadi_object(val) or not is_pickleable(val):
-                slotsstate.pop(attr, None)
-    return slotsstate
+    state: Dict[str, Any] = {attr: getattr(obj, attr, None) for attr in attributes}
+    if not fullstate:
+        for attr in list(state.keys()):
+            val = state[attr]
+            if (
+                (remove_None and val is None)
+                or is_casadi_object(val)
+                or not is_pickleable(val)
+            ):
+                state.pop(attr, None)
+    return state
 
 
 T = TypeVar("T", bound="SupportsDeepcopyAndPickle")
 
 
 class SupportsDeepcopyAndPickle:
-    """Class that defines a `__getstate__` that is compatible with both
-    `deepcopy` and `pickle`, as well as any other operation that requires the instance's
-    state.
+    """Class that defines a `__getstate__` that is compatible with both `deepcopy` and
+    `pickle`, as well as any other operation that requires the instance's state.
 
-    When pickled, use the context manager `pickleable` in order to automatically remove
-    states that cannot be pickled (e.g., CasADi objects); otherwise, use `fullstate`
-    for, e.g., `deepcopy`ing the class instance."""
-
-    _GETFULLSTATE: Optional[bool] = None
-
-    @contextmanager
-    def pickleable(self) -> Iterator[None]:
-        """Context manager that makes the class pickleable by automatically removing
-        unpickleable states (opposite of `fullstate`)."""
-        self._GETFULLSTATE = False
-        yield
-        self._GETFULLSTATE = None
-
-    @contextmanager
-    def fullstate(self) -> Iterator[None]:
-        """Context manager that makes the class return the full state without removing
-        unpickleable states (opposite of `pickleable`)."""
-        self._GETFULLSTATE = True
-        yield
-        self._GETFULLSTATE = None
+    When pickled, states that cannot be pickled (e.g., CasADi objects) are automatically
+    removed."""
 
     def copy(self: T, invalidate_caches: bool = True) -> T:
         """Creates a deepcopy of this instance.
@@ -127,8 +94,7 @@ class SupportsDeepcopyAndPickle:
         `SupportsDeepcopyAndPickle` or its subclass
             A deepcopy of this instance.
         """
-        with self.fullstate():
-            new = deepcopy(self)
+        new = deepcopy(self)
         if invalidate_caches:
             # basically do again what csnlp.util.funcs.invalidate_cache does
             for membername, member in getmembers(type(new)):
@@ -139,28 +105,33 @@ class SupportsDeepcopyAndPickle:
                         del new.__dict__[membername]
         return new
 
+    def __deepcopy__(self: T, memo: Optional[Dict[int, List[Any]]] = None) -> T:
+        """Returns a deepcopy of the object."""
+        rv = self.__reduce_ex__(4)
+        if isinstance(rv, str):
+            return self
+        assert len(rv) < 6 or rv[5] is None, "Unexpected reductor callable."
+        # overwrite the filtered state with its full version
+        fullstate = self.__getstate__(fullstate=True)
+        new_rv = (*rv[:2], fullstate, *rv[3:])
+        return _reconstruct(self, memo, *new_rv)
+
     def __getstate__(
-        self,
+        self: T,
+        fullstate: bool = False,
     ) -> Union[None, Dict[str, Any], Tuple[Optional[Dict[str, Any]], Dict[str, Any]]]:
-        """Returns the instance's state to be pickled or copied."""
+        """Returns the instance's state to be pickled/deepcopied."""
         # https://docs.python.org/3/library/pickle.html#pickle-inst
-        has_dict = hasattr(self, "__dict__")
-        has_slots = hasattr(self, "__slots__")
-        if not has_dict and not has_slots:
-            return None
-
-        if self._GETFULLSTATE is None:
-            raise RuntimeError(
-                f"Trying to get the state of {self.__class__.__name__} without using "
-                "context manager `pickleable` or `fullstate`."
-            )
-
-        if has_dict and not has_slots:
-            return _get_dict_state(self)
-
-        if has_dict:
-            return _get_dict_state(self), _get_slots_state(self)
-        return None, _get_slots_state(self)
+        dictstate = (
+            _get_dict_state(self, self.__dict__.keys(), fullstate, False)
+            if hasattr(self, "__dict__")
+            else None
+        )
+        slots = chain.from_iterable(
+            getattr(cls, "__slots__", []) for cls in type(self).__mro__
+        )
+        slotstate = _get_dict_state(self, slots, fullstate, True) if slots else None
+        return (dictstate, slotstate) if slotstate is not None else dictstate
 
 
 def save(filename: str, **data: Any) -> str:
