@@ -6,7 +6,7 @@ import numpy.typing as npt
 
 from csnlp.wrappers.wrapper import Nlp, NonRetroactiveWrapper
 
-T = TypeVar("T", cs.SX, cs.MX)
+SymType = TypeVar("SymType", cs.SX, cs.MX)
 
 
 def _n(statename: str) -> str:
@@ -14,7 +14,7 @@ def _n(statename: str) -> str:
     return f"{statename}_0"
 
 
-class Mpc(NonRetroactiveWrapper[T]):
+class Mpc(NonRetroactiveWrapper[SymType]):
     """A wrapper to easily turn the NLP scheme into an MPC controller. Most of the
     theory for MPC is taken from [1].
 
@@ -24,9 +24,22 @@ class Mpc(NonRetroactiveWrapper[T]):
         theory, computation, and design (Vol. 2). Madison, WI: Nob Hill Publishing.
     """
 
+    __slots__ = (
+        "_is_multishooting",
+        "_prediction_horizon",
+        "_control_horizon",
+        "_states",
+        "_initial_states",
+        "_actions",
+        "_actions_exp",
+        "_slacks",
+        "_disturbances",
+        "_dynamics",
+    )
+
     def __init__(
         self,
-        nlp: Nlp[T],
+        nlp: Nlp[SymType],
         prediction_horizon: int,
         control_horizon: Optional[int] = None,
         shooting: Literal["single", "multi"] = "multi",
@@ -76,12 +89,12 @@ class Mpc(NonRetroactiveWrapper[T]):
         else:
             self._control_horizon = control_horizon
 
-        self._states: Dict[str, T] = {}
-        self._initial_states: Dict[str, T] = {}
-        self._actions: Dict[str, T] = {}
-        self._actions_exp: Dict[str, T] = {}
-        self._slacks: Dict[str, T] = {}
-        self._disturbances: Dict[str, T] = {}
+        self._states: Dict[str, SymType] = {}
+        self._initial_states: Dict[str, SymType] = {}
+        self._actions: Dict[str, SymType] = {}
+        self._actions_exp: Dict[str, SymType] = {}
+        self._slacks: Dict[str, SymType] = {}
+        self._disturbances: Dict[str, SymType] = {}
         self._dynamics: cs.Function = None
 
     @property
@@ -95,27 +108,37 @@ class Mpc(NonRetroactiveWrapper[T]):
         return self._control_horizon
 
     @property
-    def states(self) -> Dict[str, T]:
+    def states(self) -> Dict[str, SymType]:
         """Gets the states of the MPC controller."""
         return self._states
 
     @property
-    def initial_states(self) -> Dict[str, T]:
+    def initial_states(self) -> Dict[str, SymType]:
         """Gets the initial states (parameters) of the MPC controller."""
         return self._initial_states
 
+    @property
+    def first_states(self) -> Dict[str, SymType]:
+        """Gets the first (along the prediction horizon) states of the controller."""
+        return {n: s[:, 0] for n, s in self._states.items()}
+
+    @property
+    def first_actions(self) -> Dict[str, SymType]:
+        """Gets the first (along the prediction horizon) actions of the controller."""
+        return {n: a[:, 0] for n, a in self._actions.items()}
+
     @property
     def ns(self) -> int:
         """Gets the number of states of the MPC controller."""
         return sum(x0.shape[0] for x0 in self._initial_states.values())
 
     @property
-    def actions(self) -> Dict[str, T]:
+    def actions(self) -> Dict[str, SymType]:
         """Gets the control actions of the MPC controller."""
         return self._actions
 
     @property
-    def actions_expanded(self) -> Dict[str, T]:
+    def actions_expanded(self) -> Dict[str, SymType]:
         """Gets the expanded control actions of the MPC controller."""
         return self._actions_exp
 
@@ -125,7 +148,7 @@ class Mpc(NonRetroactiveWrapper[T]):
         return sum(a.shape[0] for a in self._actions.values())
 
     @property
-    def slacks(self) -> Dict[str, T]:
+    def slacks(self) -> Dict[str, SymType]:
         """Gets the slack variables of the MPC controller."""
         return self._slacks
 
@@ -135,7 +158,7 @@ class Mpc(NonRetroactiveWrapper[T]):
         return sum(s.shape[0] for s in self._slacks.values())
 
     @property
-    def disturbances(self) -> Dict[str, T]:
+    def disturbances(self) -> Dict[str, SymType]:
         """Gets the disturbance parameters of the MPC controller."""
         return self._disturbances
 
@@ -159,7 +182,7 @@ class Mpc(NonRetroactiveWrapper[T]):
         size: int = 1,
         lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
         ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
-    ) -> Tuple[Optional[T], T]:
+    ) -> Tuple[Optional[SymType], SymType]:
         """Adds a state variable to the MPC controller along the whole prediction
         horizon. Automatically creates the constraint on the initial conditions for this
         state.
@@ -215,7 +238,7 @@ class Mpc(NonRetroactiveWrapper[T]):
         size: int = 1,
         lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
         ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
-    ) -> Tuple[T, T]:
+    ) -> Tuple[SymType, SymType]:
         """Adds a control action variable to the MPC controller along the whole control
         horizon. Automatically expands this action to be of the same length of the
         prediction horizon by padding with the final action.
@@ -246,7 +269,7 @@ class Mpc(NonRetroactiveWrapper[T]):
         self._actions_exp[name] = u_exp
         return u, u_exp
 
-    def disturbance(self, name: str, size: int = 1) -> T:
+    def disturbance(self, name: str, size: int = 1) -> SymType:
         """Adds a disturbance parameter to the MPC controller along the whole prediction
         horizon.
 
@@ -269,12 +292,12 @@ class Mpc(NonRetroactiveWrapper[T]):
     def constraint(
         self,
         name: str,
-        lhs: Union[T, np.ndarray, cs.DM],
+        lhs: Union[SymType, np.ndarray, cs.DM],
         op: Literal["==", ">=", "<="],
-        rhs: Union[T, np.ndarray, cs.DM],
+        rhs: Union[SymType, np.ndarray, cs.DM],
         soft: bool = False,
         simplify: bool = True,
-    ) -> Tuple[T, ...]:
+    ) -> Tuple[SymType, ...]:
         """See `Nlp.constraint` method."""
         out = self.nlp.constraint(
             name=name, lhs=lhs, op=op, rhs=rhs, soft=soft, simplify=simplify
@@ -323,18 +346,18 @@ class Mpc(NonRetroactiveWrapper[T]):
             n_out = F.n_out()
         elif n_in is None or n_out is None:
             raise ValueError(
-                "Args `n_in` and `n_out` must be manually specified since F is not a "
+                "Args `n_in` and `n_out` must be manually specified when F is not a "
                 "casadi function."
             )
-        if n_in < 2 or n_in > 3 or n_out < 1:  # type: ignore
+        if n_in is None or n_in < 2 or n_in > 3 or n_out is None or n_out < 1:
             raise ValueError(
                 "The dynamics function must accepted 2 or 3 arguments and return at "
                 f"at least 1 output; got {n_in} inputs and {n_out} outputs instead."
             )
         if self._is_multishooting:
-            self._multishooting_dynamics(F, n_in, n_out)  # type: ignore
+            self._multishooting_dynamics(F, n_in, n_out)
         else:
-            self._singleshooting_dynamics(F, n_in, n_out)  # type: ignore
+            self._singleshooting_dynamics(F, n_in, n_out)
         self._dynamics = F
 
     def _multishooting_dynamics(self, F: cs.Function, n_in: int, n_out: int) -> None:
@@ -342,10 +365,14 @@ class Mpc(NonRetroactiveWrapper[T]):
         X = cs.vertcat(*self._states.values())
         U = cs.vertcat(*self._actions_exp.values())
         if n_in < 3:
-            args_at = lambda k: (X[:, k], U[:, k])  # type: ignore # noqa: E731
+            args_at = lambda k: (X[:, k], U[:, k])  # noqa: E731
         else:
             D = cs.vertcat(*self._disturbances.values())
-            args_at = lambda k: (X[:, k], U[:, k], D[:, k])  # type: ignore # noqa: E731
+            args_at = lambda k: (  # type: ignore[return-value,assignment] # noqa: E731
+                X[:, k],
+                U[:, k],
+                D[:, k],
+            )
         xs_next = []
         for k in range(self._prediction_horizon):
             x_next = F(*args_at(k))
@@ -360,10 +387,13 @@ class Mpc(NonRetroactiveWrapper[T]):
         Xk = cs.vertcat(*self._initial_states.values())
         U = cs.vertcat(*self._actions_exp.values())
         if n_in < 3:
-            args_at = lambda k: (U[:, k],)  # type: ignore  # noqa: E731
+            args_at = lambda k: (U[:, k],)  # noqa: E731
         else:
             D = cs.vertcat(*self._disturbances.values())
-            args_at = lambda k: (U[:, k], D[:, k])  # type: ignore  # noqa: E731
+            args_at = lambda k: (  # type: ignore[return-value,assignment] # noqa: E731
+                U[:, k],
+                D[:, k],
+            )
         X = [Xk]
         for k in range(self._prediction_horizon):
             Xk = F(Xk, *args_at(k))
